package ru.dbotthepony.kstarbound.world.entities

import ru.dbotthepony.kstarbound.GlobalDefaults
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.client.StarboundClient
import ru.dbotthepony.kstarbound.defs.BaseMovementParameters
import ru.dbotthepony.kstarbound.defs.ignorePlatformCollision
import ru.dbotthepony.kstarbound.defs.restDuration
import ru.dbotthepony.kstarbound.util.MailboxExecutorService
import ru.dbotthepony.kstarbound.world.Chunk
import ru.dbotthepony.kstarbound.world.World
import ru.dbotthepony.kstarbound.world.physics.CollisionPoly
import ru.dbotthepony.kstarbound.world.physics.CollisionType
import ru.dbotthepony.kstarbound.world.physics.Poly
import ru.dbotthepony.kvector.util.linearInterpolation
import ru.dbotthepony.kvector.util2d.AABB
import ru.dbotthepony.kvector.vector.RGBAColor
import ru.dbotthepony.kvector.vector.Vector2d
import ru.dbotthepony.kvector.vector.times
import java.util.EnumSet
import kotlin.concurrent.withLock
import kotlin.math.PI
import kotlin.math.absoluteValue
import kotlin.math.acos

abstract class Entity(val world: World<*, *>) {
	var chunk: Chunk<*, *>? = null
		set(value) {
			if (!isSpawned) {
				throw IllegalStateException("Trying to set chunk this entity belong to before spawning in world")
			} else if (isRemoved) {
				throw IllegalStateException("This entity was removed")
			} else if (value == field) {
				return
			}

			val chunkPos = world.chunkFromCell(position)

			if (value != null && chunkPos != value.pos) {
				throw IllegalStateException("Set proper position before setting chunk this Entity belongs to (expected chunk $chunkPos, got chunk ${value.pos})")
			}

			val oldChunk = field
			field = value

			world.lock.withLock {
				if (oldChunk == null && value != null) {
					world.orphanedEntities.remove(this)
					value.addEntity(this)
				} else if (oldChunk != null && value == null) {
					world.orphanedEntities.add(this)
					oldChunk.removeEntity(this)
				} else if (oldChunk != null && value != null) {
					value.transferEntity(this, oldChunk)
				}
			}
		}

	var position = Vector2d()
		set(value) {
			if (field == value)
				return

			val old = field
			field = Vector2d(world.x.cell(value.x), world.y.cell(value.y))
			physicsSleepTicks = 0

			if (isSpawned && !isRemoved) {
				val oldChunkPos = world.chunkFromCell(old)
				val newChunkPos = world.chunkFromCell(field)

				if (oldChunkPos != newChunkPos) {
					chunk = world.chunkMap[newChunkPos]
				}
			}
		}

	var velocity = Vector2d.ZERO
		set(value) {
			field = value
			physicsSleepTicks = 0
		}

	// Movement variables
	var isZeroGravity = false
		private set

	var isOnGround = false
		private set
	var isColliding = false
		private set
	var isCollisionStuck = false
		private set
	var isCollidingWithNull = false
		private set
	var stickingDirection: Double? = null
		private set
	var surfaceSlope = Vector2d.ZERO
		private set
	var surfaceVelocity = Vector2d.ZERO
		private set
	var collisionCorrection = Vector2d.ZERO
		private set
	var liquidPercentage = 0.0
		private set

	// Movement parameters
	open val movementParameters: BaseMovementParameters = GlobalDefaults.movementParameters

	var gravityMultiplier = 1.0
	var isGravityDisabled = false

	var mass = 1.0
		set(value) {
			require(value > 0.0) { "Invalid mass: $value" }
			field = value
		}

	var physicsSleepTicks = 0
	val mailbox = MailboxExecutorService(world.mailbox.thread)

	protected val hitboxes = ArrayList<Poly>()
	protected val collisionFilter: EnumSet<CollisionType> = EnumSet.of(CollisionType.NONE)

	/**
	 * true - whitelist, false - blacklist
	 */
	protected var collisionFilterMode = false

	/**
	 * Whenever is this entity spawned in world ([spawn] called).
	 * Doesn't mean entity still exists in world, check it with [isRemoved]
	 */
	var isSpawned = false
		private set

	/**
	 * Whenever is this entity was removed from world ([remove] called).
	 */
	var isRemoved = false
		private set

	open fun spawn() {
		if (isSpawned)
			throw IllegalStateException("Already spawned")

		isSpawned = true

		world.mailbox.execute {
			world.entities.add(this)
			chunk = world.chunkMap[world.chunkFromCell(position)]

			if (chunk == null) {
				world.orphanedEntities.add(this)
			}
		}
	}

	open fun remove() {
		if (isRemoved)
			throw IllegalStateException("Already removed")

		isRemoved = true
		mailbox.shutdownNow()

		if (isSpawned) {
			world.mailbox.execute {
				world.entities.remove(this)
				chunk?.removeEntity(this)
			}
		}
	}

	/**
	 * this function is executed sequentially
	 */
	open fun thinkShared() {
		mailbox.executeQueuedTasks()
	}

	open fun thinkClient() {

	}

	open fun thinkServer() {

	}

	fun updateLiquidPercentage() {

	}

	fun updateForceRegions() {

	}

	fun determineGravity(): Vector2d {
		if (isZeroGravity || isGravityDisabled)
			return Vector2d.ZERO

		return world.gravityAt(position)
	}

	/**
	 * this function is executed in parallel
	 */
	// TODO: Ghost collisions occur, where objects trip on edges
	open fun move() {
		isZeroGravity = isGravityDisabled || gravityMultiplier == 0.0 || determineGravity().lengthSquared == 0.0

		if (!isZeroGravity)
			velocity += world.gravity * Starbound.TICK_TIME_ADVANCE

		movementParameters.speedLimit?.let {
			if (velocity.length > it)
				velocity = velocity.unitVector * it
		}

		// TODO: Here: moving platforms sticky code

		if (hitboxes.isEmpty() || movementParameters.collisionEnabled != true) {
			position += velocity * Starbound.TICK_TIME_ADVANCE
			surfaceSlope = Vector2d.POSITIVE_Y
			surfaceVelocity = Vector2d.ZERO
			isOnGround = false
			stickingDirection = null
			isColliding = false
			isCollidingWithNull = false
			isCollisionStuck = false
			return
		}

		var steps = 1

		movementParameters.maxMovementPerStep?.let {
			steps = (velocity.length * Starbound.TICK_TIME_ADVANCE / it).toInt() + 1
		}

		var relativeVelocity = if (physicsSleepTicks > 0) {
			physicsSleepTicks--
			Vector2d.ZERO
		} else {
			velocity
		}

		val originalMovement = relativeVelocity * Starbound.TICK_TIME_ADVANCE
		surfaceSlope = Vector2d.POSITIVE_Y
		// TODO: Here: moving platforms sticky code

		val dt = Starbound.TICK_TIME_ADVANCE / steps

		for (step in 0 until steps) {
			val velocityMagnitude = relativeVelocity.length
			val velocityDirection = relativeVelocity / velocityMagnitude
			val movement = relativeVelocity * dt

			val ignorePlatforms = movementParameters.ignorePlatformCollision || relativeVelocity.y > 0.0
			val maximumCorrection = movementParameters.maximumCorrection ?: 0.0
			val maximumPlatformCorrection = (movementParameters.maximumPlatformCorrection ?: Double.POSITIVE_INFINITY) +
				(movementParameters.maximumPlatformCorrectionVelocityFactor ?: 0.0) * velocityMagnitude

			val localHitboxes = hitboxes.map { it + position }
			val aabb = localHitboxes.stream().map { it.aabb }.reduce(AABB::combine).get()
			var queryBounds = aabb.enlarge(maximumCorrection, maximumCorrection)
			queryBounds = queryBounds.combine(queryBounds + movement)

			val polies = world.queryCollisions(queryBounds).filter {
				if (collisionFilterMode)
					it.type in collisionFilter
				else
					it.type !in collisionFilter
			}

			val results = ArrayList<CollisionResult>(localHitboxes.size)

			for (hitbox in localHitboxes) {
				results.add(collisionSweep(hitbox, polies, movement, ignorePlatforms, (movementParameters.enableSurfaceSlopeCorrection ?: false) && !isZeroGravity, maximumCorrection, maximumPlatformCorrection, aabb.centre))
			}

			val result = results.minOrNull()!!
			position += result.movement

			if (result.collisionType == CollisionType.NULL) {
				isCollidingWithNull = true
				break
			} else {
				isCollidingWithNull = false
			}

			val correction = result.correction
			val normCorrection = correction.unitVector
			surfaceSlope = result.groundSlope
			collisionCorrection = result.correction
			isColliding = correction != Vector2d.ZERO || result.isStuck
			isOnGround = !isZeroGravity && result.isOnGround
			isCollisionStuck = result.isStuck

			// If we have collided, apply either sticky or normal (bouncing) collision physics
			if (correction != Vector2d.ZERO) {
				if (movementParameters.stickyCollision == true && result.collisionType !== CollisionType.SLIPPERY) {
					// When sticking, cancel all velocity and apply stickyForce in the
					// opposite of the direction of collision correction.
					relativeVelocity = -normCorrection * (movementParameters.stickyForce ?: 0.0) / mass * dt
					stickingDirection = -normCorrection.toAngle()
					break
				} else {
					stickingDirection = null
					val correctionDirection = correction.unitVector

					if (movementParameters.bounceFactor != null && movementParameters.bounceFactor != 0.0) {
						val adjustment = correctionDirection * (velocityMagnitude * (correctionDirection * -velocityDirection))
						relativeVelocity += adjustment + movementParameters.bounceFactor!! * adjustment

						if (movementParameters.stopOnFirstBounce == true) {
							// When bouncing, stop integrating at the moment of bounce.  This
							// prevents the frame of contact from being missed due to multiple
							// iterations per frame.
							break
						}
					} else {
						// Only adjust the velocity to the extent that the collision was
						// caused by the velocity in each axis, to eliminate collision
						// induced velocity in a platformery way (each axis considered
						// independently).

						if (relativeVelocity.x < 0.0 && correction.x > 0.0)
							relativeVelocity = relativeVelocity.copy(x = (relativeVelocity.x + correction.x / Starbound.TICK_TIME_ADVANCE).coerceAtMost(0.0))
						else if (relativeVelocity.x > 0.0 && correction.x < 0.0)
							relativeVelocity = relativeVelocity.copy(x = (relativeVelocity.x + correction.x / Starbound.TICK_TIME_ADVANCE).coerceAtLeast(0.0))

						if (relativeVelocity.y < 0.0 && correction.y > 0.0)
							relativeVelocity = relativeVelocity.copy(y = (relativeVelocity.y + correction.y / Starbound.TICK_TIME_ADVANCE).coerceAtMost(0.0))
						else if (relativeVelocity.y > 0.0 && correction.y < 0.0)
							relativeVelocity = relativeVelocity.copy(y = (relativeVelocity.y + correction.y / Starbound.TICK_TIME_ADVANCE).coerceAtLeast(0.0))
					}
				}
			}
		}

		var newVelocity = relativeVelocity

		updateLiquidPercentage()

		// TODO: sticky collision update

		// In order to make control work accurately, passive forces need to be
		// applied to velocity *after* integrating.  This prevents control from
		// having to account for one timestep of passive forces in order to result
		// in the correct controlled movement.
		if (!isZeroGravity && stickingDirection == null) {
			val buoyancy = (movementParameters.liquidBuoyancy ?: 0.0).coerceIn(0.0, 1.0) + liquidPercentage + (movementParameters.airBuoyancy ?: 0.0).coerceIn(0.0, 1.0) * (1.0 - liquidPercentage)
			val gravity = determineGravity() * (movementParameters.gravityMultiplier ?: 1.0) * (1.0 - buoyancy)
			var environmentVelocity = gravity * Starbound.TICK_TIME_ADVANCE

			if (isOnGround && (movementParameters.slopeSlidingFactor ?: 0.0) != 0.0 && surfaceSlope != Vector2d.ZERO)
				environmentVelocity += surfaceSlope * (surfaceSlope.x * surfaceSlope.y) * (movementParameters.slopeSlidingFactor ?: 0.0)

			newVelocity += environmentVelocity
		}

		// If original movement was entirely (almost) in the direction of gravity
		// and was entirely (almost) cancelled by collision correction, put the
		// entity into rest for restDuration
		if (
			physicsSleepTicks == 0 &&
			originalMovement.dot(determineGravity()) in 0.99 .. 1.01 &&
			collisionCorrection.dot(determineGravity()) in -1.01 .. -0.99
		) {
			physicsSleepTicks = movementParameters.restDuration
		}

		if (movementParameters.frictionEnabled == true) {
			var refVel = Vector2d.ZERO
			var friction = liquidPercentage * (movementParameters.liquidFriction ?: 0.0).coerceIn(0.0, 1.0) + (1.0 - liquidPercentage) * (movementParameters.airFriction ?: 0.0).coerceIn(0.0, 1.0)

			if (isOnGround) {
				friction = friction.coerceAtLeast(movementParameters.groundFriction ?: 0.0)
				refVel = surfaceVelocity
			}

			// The equation for friction here is effectively:
			// frictionForce = friction * (refVel - velocity)
			// but it is applied here as a multiplicative factor from [0, 1] so it does
			// not induce oscillation at very high friction and so it cannot be
			// negative.
			val frictionFactor = (friction / mass * Starbound.TICK_TIME_ADVANCE).coerceIn(0.0, 1.0)
			newVelocity = linearInterpolation(frictionFactor, newVelocity, refVel)
		}

		velocity = newVelocity
		updateForceRegions()
	}

	protected data class CollisionSeparation(
		var correction: Vector2d = Vector2d.ZERO,
		var solutionFound: Boolean = false,
		var collisionType: CollisionType = CollisionType.NONE,
		var axis: Vector2d = Vector2d.POSITIVE_Y,
		var movingCollisionId: Int? = null, // TODO
	)

	protected data class CollisionResult(
		val movement: Vector2d = Vector2d.ZERO,
		val correction: Vector2d = Vector2d.ZERO,
		var movingCollisionId: Int? = null, // TODO
		var isStuck: Boolean = false,
		var isOnGround: Boolean = false,
		var groundSlope: Vector2d = Vector2d.ZERO,
		var collisionType: CollisionType = CollisionType.NULL,
	) : Comparable<CollisionResult> {
		override fun compareTo(other: CollisionResult): Int {
			return movement.lengthSquared.compareTo(other.movement.lengthSquared)
		}
	}

	protected fun collisionSweep(
		body: Poly, staticBodies: List<CollisionPoly>,
		movement: Vector2d, ignorePlatforms: Boolean,
		slopeCorrection: Boolean, maximumCorrection: Double,
		maximumPlatformCorrection: Double, sortCenter: Vector2d
	): CollisionResult {
		val translatedBody = body + movement
		var checkBody = translatedBody
		var maxCollided = CollisionType.NONE

		var separation = CollisionSeparation()
		var totalCorrection = Vector2d.ZERO
		var movingCollisionId: Int? = null

		val sorted = staticBodies.stream()
			.map { it to (it.poly.aabb.centre - sortCenter).lengthSquared }
			.sorted { o1, o2 -> o1.second.compareTo(o2.second) }
			.map { it.first }
			.toList()

		if (slopeCorrection) {
			// Starbound: First try separating with our ground sliding cheat.
			separation = collisionSeparate(checkBody, sorted, ignorePlatforms, maximumPlatformCorrection, true, SEPARATION_TOLERANCE)
			totalCorrection += separation.correction
			checkBody += separation.correction
			maxCollided = maxCollided.maxOf(separation.collisionType)
			movingCollisionId = separation.movingCollisionId

			val upwardResult = movement + separation.correction
			val upwardMagnitude = upwardResult.length
			val upwardUnit = upwardResult / upwardMagnitude
			// Starbound: Angle off of horizontal (minimum of either direction)
			val horizontalAngle = acos(Vector2d.POSITIVE_X.dot(upwardUnit)).coerceAtMost(acos(Vector2d.NEGATIVE_X.dot(upwardUnit)))

			// Starbound: We need to make sure that even if we found a solution with the sliding
			// Starbound: cheat, we are not beyond the angle and correction limits for the ground
			// Starbound: cheat correction.
			separation.solutionFound = separation.solutionFound && (upwardMagnitude < 0.2 || horizontalAngle < PI / 3.0) && totalCorrection.length <= maximumCorrection

			// KStarbound: if we got pushed into world geometry, then consider slide cheat didn't find a solution
			if (separation.solutionFound) {
				separation.solutionFound = staticBodies.all { it.poly.intersect(checkBody).let { it == null || it.penetration.absoluteValue <= SEPARATION_TOLERANCE } }
			}
		}

		if (!separation.solutionFound) {
			checkBody = translatedBody
			totalCorrection = Vector2d.ZERO
			movingCollisionId = null

			for (i in 0 until SEPARATION_STEPS) {
				separation = collisionSeparate(checkBody, sorted, ignorePlatforms, maximumPlatformCorrection, false, SEPARATION_TOLERANCE)
				totalCorrection += separation.correction
				checkBody += separation.correction
				maxCollided = maxCollided.maxOf(separation.collisionType)

				if (totalCorrection.length >= maximumCorrection) {
					separation.solutionFound = false
					break
				} else if (separation.solutionFound) {
					break
				}
			}
		}

		if (!separation.solutionFound && movement != Vector2d.ZERO) {
			checkBody = body
			totalCorrection = -movement

			for (i in 0 until SEPARATION_STEPS) {
				separation = collisionSeparate(checkBody, sorted, true, maximumPlatformCorrection, false, SEPARATION_TOLERANCE)
				totalCorrection += separation.correction
				checkBody += separation.correction
				maxCollided = maxCollided.maxOf(separation.collisionType)

				if (totalCorrection.length >= maximumCorrection) {
					separation.solutionFound = false
					break
				} else if (separation.solutionFound) {
					break
				}
			}
		}

		if (separation.solutionFound) {
			val result = CollisionResult(
				movement = movement + totalCorrection,
				correction = totalCorrection,
				isStuck = false,
				isOnGround = -totalCorrection.dot(determineGravity()) > SEPARATION_TOLERANCE,
				movingCollisionId = movingCollisionId,
				collisionType = maxCollided,
				// groundSlope = Vector2d.POSITIVE_Y,
				groundSlope = separation.axis
			)

			// TODO: what they wanted to achieve with this?
			/*if (result.isOnGround) {
				// If we are on the ground and need to find the ground slope, look for a
			    // vertex on the body being moved that is touching an edge of one of the
			    // collision polys.  We only want a slope to be produced from an edge of
			    // colision geometry, not an edge of the colliding body.  Pick the
			    // touching edge that is the most horizontally overlapped with the
			    // geometry, rather than off to the side.
				var maxSideHorizontalOverlap = 0.0
				var touchingBounds = checkBody.aabb.enlarge(SEPARATION_TOLERANCE, SEPARATION_TOLERANCE)

				for (poly in staticBodies) {
					for (edge in poly.poly.edges) {

					}
				}
			}*/

			return result
		} else {
			return CollisionResult(Vector2d.ZERO, -movement, null, true, true, Vector2d.POSITIVE_Y, maxCollided)
		}
	}

	protected fun collisionSeparate(
		poly: Poly, staticBodies: List<CollisionPoly>,
		ignorePlatforms: Boolean, maximumPlatformCorrection: Double,
		upward: Boolean, separationTolerance: Double
	): CollisionSeparation {
		val separation = CollisionSeparation()
		var intersects = false
		var correctedPoly = poly

		for (body in staticBodies) {
			if (ignorePlatforms && body.type === CollisionType.PLATFORM)
				continue

			var result = if (upward)
				correctedPoly.intersect(body.poly, Vector2d.POSITIVE_Y, false)
			else if (body.type == CollisionType.PLATFORM)
				correctedPoly.intersect(body.poly, Vector2d.POSITIVE_Y, true)
			else
				correctedPoly.intersect(body.poly)

			if (body.type === CollisionType.PLATFORM && result != null && (result.penetration <= 0.0 || result.penetration > maximumPlatformCorrection))
				result = null

			if (result != null) {
				intersects = true
				correctedPoly += result.vector
				separation.correction += result.vector
				separation.collisionType = separation.collisionType.maxOf(body.type)
			}
		}

		separation.solutionFound = true

		if (intersects) {
			for (body in staticBodies) {
				if (body.type === CollisionType.PLATFORM)
					continue

				val result = correctedPoly.intersect(body.poly)

				if (result != null && result.penetration > separationTolerance) {
					separation.collisionType = separation.collisionType.maxOf(body.type)
					separation.solutionFound = false
					break
				}
			}
		}

		return separation
	}

	protected open fun onTouch(velocity: Vector2d, normal: Vector2d, poly: CollisionPoly) {

	}

	open fun render(client: StarboundClient = StarboundClient.current()) {
		hitboxes.forEach { (it + position).render(client) }

		world.queryCollisions(
			hitboxes.stream().map { (it + position).aabb }.reduce(AABB::combine).get().enlarge(2.0, 2.0)
		).filter {
			if (collisionFilterMode)
				it.type in collisionFilter
			else
				it.type !in collisionFilter
		}.forEach { it.poly.render(client, BLOCK_COLLISION_COLOR) }
	}

	open var maxHealth = 0.0
	open var health = 0.0

	open fun hurt(amount: Double): Boolean {
		return false
	}

	companion object {
		const val PHYSICS_TICKS_UNTIL_SLEEP = 16
		val BLOCK_COLLISION_COLOR = RGBAColor(65, 179, 217)
		const val SEPARATION_STEPS = 3
		const val SEPARATION_TOLERANCE = 0.001
	}
}
